unit interpolation1d;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, StdCtrls, PairSplitter, Spin, guitabbase,
  xmlUtil, Graph2D, DataClasses, dialogs, mathconfig, splines,
  multidim, Graph2DFunctions, poly, math, numericAnalysis;

type

  { TPointItem }

  TPointItem = class(TDataItem)
   private
    labelX, labelY, labelD : TLabel;
    editX, editY : TEdit;
    spinEditD : TSpinEdit;
    movePoint : TGraph2D.TMovePoint;
    procedure resize(Sender:TObject); reintroduce;
    procedure doChange(Sender:TObject);
    procedure pointMoved(Sender:TObject);
    procedure removeClick(Sender:TObject);
   protected
    procedure afterInsertion; override;
   public
    constructor create(theXml:XML.TElement); override;
    destructor destroy; override;
  end;

  TExternalButDataClass = class(TDataClass)
   public
    property addButton : TButton read addBut;
    property scrollBox : TScrollBox read box;
  end;

  { TInterpolation1Dim }
  TInterpolation1Dim = class(TGuiTab)
     Label1: TLabel;
     Label2: TLabel;
     Label3: TLabel;
     PairSplitter1: TPairSplitter;
     PairSplitter2: TPairSplitter;
     PairSplitterSide1: TPairSplitterSide;
     PairSplitterSide2: TPairSplitterSide;
     PairSplitterSide3: TPairSplitterSide;
     PairSplitterSide4: TPairSplitterSide;
     ComboBox1: TComboBox;
     GroupBox1: TGroupBox;
     Edit1: TEdit;
     Edit2: TEdit;
     Edit3: TEdit;
     Label5: TLabel;
     CheckBox1: TCheckBox;
     CheckBox2: TCheckBox;
     CheckBox3: TCheckBox;
     ComboBox2: TComboBox;
     Label7: TLabel;
     Edit4: TEdit;
     Label8: TLabel;
     Edit5: TEdit;
     Label9: TLabel;
     procedure BordaryChanged(Sender: TObject);
     procedure InterpolationChange(Sender: TObject);
     procedure UpdateIntegral(Sender: TObject);
   type
    SMD = specialize Multidimensional<MD1d, MD1d>;
   procedure visibilityChange(Sender: TObject);
   private
    integralX1, integralX2 : Precision;
    graph : TGraph2D;
    graphFkt : TGraphFunction;
    graphFktD : TGraphFunction;
    graphFktDD : TGraphFunction;
    points : TExternalButDataClass;
    bordary : Spline.TCubicC2.TBordaryCondition;
    periodicBorders : array [0..1] of TGraph2D.TBar;
    fa : Precision;
    fb : Precision;
    class var instanceID : Cardinal;
    procedure interpolate;
   public
    constructor create(theOwner:TControl; __xmlEl:XML.TElement); override;
    destructor destroy; override;
  end;

implementation

{$R *.lfm}

{ TPointItem }

constructor TPointItem.create(theXml: XML.TElement);
begin
  inherited create(theXml);
  butMod.Visible := false;
  butRem.onClick := @removeClick;
  labelX := TLabel.create(self);
  labelX.parent := self;
  labelX.Caption := 'x: ';
  labelX.anchorParallel(akLeft, 0, self);
  editX := TEdit.create(self);
  editX.Parent := self;
  editX.anchorParallel(akTop, 0, self);
  editX.text := '0';
  labelX.AnchorVerticalCenterTo(editX);
  editX.anchorToNeighbour(akLeft, 0, labelX);
  labelY := TLabel.Create(self);
  labelY.parent := self;
  labelY.anchorVerticalCenterTo(editX);
  labelY.anchorToNeighbour(akLeft, 8, editX);
  labelY.caption := 'y: ';
  editY := TEdit.create(self);
  editY.parent := self;
  editY.anchorParallel(akTop, 0, self);
  editY.anchorToNeighbour(akLeft, 0, labelY);
  editY.text := '0';
  spinEditD := TSpinEdit.create(self);
  spinEditD.Parent := self;
  spinEditD.AnchorParallel(akTop, 0, self);
  spinEditD.AnchorToNeighbour(akRight, 3, butRem);
  spinEditD.Anchors := [akTop, akRight];
  labelD := TLabel.create(self);
  labelD.Parent := self;
  labelD.AnchorVerticalCenterTo(editX);
  labelD.AnchorToNeighbour(akRight, 0, spinEditD);
  labelD.Anchors := [akTop, akRight];
  labelD.Caption := 'd: ';
  editY.AnchorToNeighbour(akRight, 8, labelD);
  self.autoSize := true;
  self.onResize := @resize;
  movePoint := nil; // gets created in afterIntertion
end;

procedure TPointItem.afterInsertion;
begin
  inherited afterInsertion;
  movePoint := TGraph2D.TMovePoint.Create(TInterpolation1Dim(parent.owner).graph);
  movePoint.movable := true;
  movePoint.onMove := @pointMoved;
  TInterpolation1Dim(parent.owner).graph.Repaint;
  editX.Text := xmlEl.getAttribute('x');
  editY.Text := xmlEl.getAttribute('y');
  spinEditD.Value := StrToIntDef(xmlEl.getAttribute('d'), 0);
  editX.OnChange := @doChange;
  editY.OnChange := @doChange;
  spinEditD.OnChange := @doChange;
  doChange(nil);
end;

procedure TPointItem.resize(Sender: TObject);
begin
  editX.Width := (editX.Width + editY.width) div 2;
end;

procedure TPointItem.pointMoved(Sender: TObject);
begin
  editX.OnChange := nil;
  editY.OnChange := nil;
  editX.Text := FloatToStr(movePoint.x, mathFloatFormat);
  editY.Text := FloatToStr(movePoint.y, mathFloatFormat);
  editX.OnChange := @doChange;
  editY.OnChange := @doChange;
  doChange(sender);
end;

procedure TPointItem.doChange(Sender: TObject);
var x, y : Precision;
begin
  editX.color := $FFFFFF;
  editY.color := $FFFFFF;
  try
    x := StrToFloat(editX.Text, mathFloatFormat);
  except
    editX.color := $AAAAFF;
  end;
  try
    y := StrToFloat(editY.Text, mathFloatFormat);
  except
    editY.color := $AAAAFF;
  end;
  if (editX.color = $AAAAFF) or (editY.color = $AAAAFF) then exit;
  if (sender <> movePoint) then begin
    movePoint.onMove := nil;
    movePoint.x := x;
    movePoint.y := y;
    movePoint.onMove := @pointMoved;
  end;
  xmlEl.setAttribute('x', editX.Text);
  xmlEl.setAttribute('y', editY.Text);
  xmlEl.setATtribute('d', IntToStr(spinEditD.Value));
  TInterpolation1Dim(parent.owner).interpolate;
end;

procedure TPointItem.removeClick(Sender:TObject);
var f : TInterpolation1Dim;
begin
  f := TInterpolation1Dim(parent.owner);
  removeAndDestroy(sender);
  movePoint.visible := false;
  f.interpolate;
end;

destructor TPointItem.destroy;
begin
  movePoint.destroy;
  inherited;
end;



{ TInterpolation1Dim }

constructor TInterpolation1Dim.create(theOwner: TControl; __xmlEl:XML.TElement);
var el : XML.TElement;
begin
  inherited create(theOwner, __xmlEl);
  name := 'Interpolation1Dim' + inttostr(instanceID);
  instanceID += 1;
  graph := TGraph2D.Create(self);
  graph.parent := PairSplitterSide1;
  graph.anchorParallel(akTop, 0, PairSplitterSide1);
  graph.anchorParallel(akLeft, 0, PairSplitterSide1);
  graph.anchorParallel(akRight, 0, PairSplitterSide1);
  graph.anchorParallel(akBottom, 0, PairSplitterSide1);
  periodicBorders[0] := TGraph2D.TBar.create(graph, VERTICAL);
  periodicBorders[1] := TGraph2D.TBar.create(graph, VERTICAL);
  graphFkt := TGraphFunction.create(graph);
  graphFktD := TGraphFunction.create(graph);
  graphFktD.style := DASHED;
  graphFktDD := TGraphFunction.create(graph);
  graphFktDD.style := DOTTED;
  Label5.caption := '';
  points := TExternalButDataClass.create(self);
  points.parent := PairSplitterSide3;
  points.AnchorToNeighbour(akTop, 0, Label1);
  points.AnchorParallel(akLeft, 0, PairSplitterSide3);
  points.AnchorParallel(akRight, 0, PairSplitterSide3);
  points.AnchorParallel(akBottom, 0, PairSplitterSide3);
  points.addItemType('POINT', TPointItem);
  points.defaultTagName := 'POINT';
  el := xmlEl.getElementsByTagName('POINTS', 0);
  if (el = nil) then begin
    el := XML.TElement.Create('POINTS');
    xmlEl.appendChild(el);
  end;
  points.xmlEl := el;
  points.addButton.Parent := PairSplitterSide3;
  points.addButton.anchorParallel(akTop, 0, PairSplitterSide3);
  points.addButton.anchorParallel(akRight, 0, PairSplitterSide3);
  points.addButton.anchorToNeighbour(akBottom, 0, points);
  points.addButton.Anchors := [akTop, akRight, akBottom];
  points.addButton.caption := 'Add fixed point';
  points.addButton.autoSize := true;
  points.scrollBox.anchorParallel(akBottom, 3, points);
  points.deleteQuestion := '';
  InterpolationChange(self);
end;


procedure TInterpolation1Dim.InterpolationChange(Sender: TObject);
type TLin = specialize Spline.TLinear<MD1d, MD1d>;
begin
  if (graphFktDD.fkt <> nil) then graphFktDD.fkt.asObject.destroy;
  if (graphFktD.fkt <> nil) then graphFktD.fkt.asObject.destroy;
  if (graphFkt.fkt <> nil) then graphFkt.fkt.asObject.destroy;
  graphFkt.fkt := nil;
  graphFktD.fkt := nil;
  graphFktDD.fkt := nil;
  GroupBox1.visible := false;
  case (ComboBox1.itemIndex) of
   0 : graphFkt.fkt := TLin.create;
   1 : begin
     graphFkt.fkt := Spline.TCubicC2.create;
     GroupBox1.visible := true;
   end;
   2 : begin
     graphFkt.fkt := TPolynom.create(0) as Analysis.IFunction;
   end;
  end;
  interpolate;
end;

procedure TInterpolation1Dim.UpdateIntegral(Sender: TObject);
var integral : Analysis.IFunction;
begin
  Label9.Caption := ' = ';
  if (Sender = Edit4) or (Sender = Edit5) then begin
    integralX1 := StrToFloatDef(Edit4.Text, NaN, mathFloatFormat);
    integralX2 := StrToFloatDef(Edit5.Text, NaN, mathFloatFormat);
  end;
  if (isNaN(integralX1)) or (isNaN(integralX2)) or (graphFkt.fkt = nil) or (not (graphFkt.fkt.asObject is Analysis.TNumericFunction)) then exit;
  integral := Analysis.TNumericFunction(graphFkt.fkt.asObject).integrate(0);
  if (integral = nil) then exit;
  Label9.Caption := ' = '+floattostr(integral.Calc(integralX2) - integral.Calc(integralX1));
  integral.asObject.Destroy;
end;

procedure TInterpolation1Dim.BordaryChanged(Sender: TObject);
begin
  Edit1.enabled := ComboBox2.itemIndex = 1;
  Edit2.enabled := ComboBox2.itemIndex = 1;
  fa := StrToFloatDef(Edit1.Text, NaN, mathFloatFormat);
  fb := StrToFloatDef(Edit2.Text, NaN, mathFloatFormat);
  if (isNaN(fa)) then Edit1.color := $AAAAFF else Edit1.color := $FFFFFF;
  if (isNaN(fb)) then Edit2.color := $AAAAFF else Edit2.color := $FFFFFF;
  case (ComboBox2.itemIndex) of
   0 : bordary := NATURAL;
   1 : bordary := HERMITE;
   2 : bordary := PERIODICAL;
   3 : bordary := NOT_A_KNOT;
  end;
  InterpolationChange(Sender);
end;

procedure TInterpolation1Dim.interpolate;
procedure splineinterpolation;
var
  buffer : SMD.TBuffer;
  it : TPointItem;
begin
  it := TPointItem(points.first);
  buffer := SMD.TBuffer.create;
  while (it <> nil) do begin
    if (not buffer.setPoint(it.movePoint.x, it.movePoint.y)) then begin
      Label5.Caption := 'Point at x='+floattostr(it.movePoint.x)+' already exists!';
      buffer.destroy;
      exit;
    end;
    it := TPointItem(it.next);
  end;
  if (buffer.grid^.laneCount[0] = 0) then begin
    buffer.destroy;
    exit;
  end;
  if (bordary = PERIODICAL) then begin
    periodicBorders[0].position := buffer.grid^[0, 0];
    periodicBorders[0].visible := true;
    periodicBorders[1].position := buffer.grid^[0, buffer.grid^.laneCount[0]-1];
    periodicBorders[1].visible := true;
  end;
  try
    if (graphFkt.fkt.asObject is Spline.TCubicC2) then
      (graphFkt.fkt.asObject as Spline.TCubicC2).computeFor(buffer, buffer.grid^, bordary, fa, fb)
    else
      (graphFkt.fkt.asObject as SMD.IInterpolation).computeFor(buffer, buffer.grid^);
  except
    on E: Exception do
      Label5.Caption := E.message;
  end;
  buffer.destroy;
end;

procedure polynominterpolation;
var
  info : TPolynom.TInterpolateInfo;
  i : integer;
  it : TPointItem;
begin
  setLength(info, points.count);
  i := 0;
  it := TPointItem(points.first);
  while (it <> nil) do begin
    info[i].x := it.movePoint.x;
    info[i].y := it.movePoint.y;
    info[i].d := it.spinEditD.value;
    i += 1;
    it := TPointItem(it.next);
  end;
  try
    graphFkt.fkt.asObject.destroy;
    graphFkt.fkt := TPolynom.hermiteInterpolation(info) as Analysis.IFunction;
  except
    on E: Exception do begin
      Label5.Caption := E.message;
      graphFkt.fkt := TPolynom.create(0) as Analysis.IFunction;
    end;
  end;
  UpdateIntegral(nil);
end;

begin
  Label5.Caption := '';
  periodicBorders[0].visible := false;
  periodicBorders[1].visible := false;
  if (graphFkt.fkt = nil) then exit;
  if (graphFkt.fkt.asObject is TPolynom) then polynominterpolation
  else if (graphFkt.fkt.asObject is SMD.IInterpolation) then splineinterpolation
  else raise Exception.Create('Cannot interpolate!');
  if (graphFkt.fkt <> nil) then Edit3.text := graphFkt.fkt.toString;
  if (graphFktD.fkt <> nil) then graphFktD.fkt.asObject.destroy;
  if (graphFktDD.fkt <> nil) then graphFktDD.fkt.asObject.destroy;
  graphFktD.fkt := nil;
  graphFktDD.fkt := nil;
  if (graphFkt.fkt.asObject is Analysis.IFunction) then begin
    graphFktD.fkt := (graphFkt.fkt.asObject as Analysis.IFunction).derivate;
    graphFktDD.fkt := (graphFktD.fkt.asObject as Analysis.IFunction).derivate;
  end;
  visibilityChange(nil);
end;

procedure TInterpolation1Dim.visibilityChange(Sender: TObject);
begin
  graphFkt.visible := CheckBox1.checked;
  graphFktD.visible := CheckBox2.checked;
  graphFktDD.visible := CheckBox3.checked;
  if (assigned(self.parent)) then graph.doPaint(self);
end;

destructor TInterpolation1Dim.destroy;
begin
  if (graphFktDD.fkt <> nil) then graphFktDD.fkt.asObject.destroy;
  if (graphFktD.fkt <> nil) then graphFktD.fkt.asObject.destroy;
  if (graphFkt.fkt <> nil) then graphFkt.fkt.asObject.destroy;
  inherited destroy;
end;




initialization
TInterpolation1Dim.instanceID := 1;
TGuiTab.addAvaillable(
  '1D interpolation',
  'Allows to apply several interpolation techniques to a onedimensional set of points.',
  TInterpolation1Dim);

end.

